# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module jsonapi/conversions deals with JSON API conversions."""

import re
from uuid import UUID

RE_PASCAL_TO_SNAKE_1 = re.compile('(.)([A-Z][a-z]+)')
RE_PASCAL_TO_SNAKE_2 = re.compile('([a-z0-9])([A-Z])')


def uuid_name_to_number(uuid_name):
    """Converts uuid name to number."""
    return UUID(uuid_name)


def id_name_to_number(id_name):
    """Converts id name to number."""
    if '-' in id_name:
        return uuid_name_to_number(id_name)
    return UUID(int=int(id_name))


def id_number_to_name(id_number):
    """Converts id number to name."""
    if id_number.int <= 0xffffffffffff:
        return str(id_number.int)
    return str(id_number)


def pascal_to_snake(name):
    """Converts class name to url name."""

    temp = RE_PASCAL_TO_SNAKE_1.sub(r'\1_\2', name)
    return RE_PASCAL_TO_SNAKE_2.sub(r'\1_\2', temp).lower()



# Copyright © 2019-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module log provides logging infrastructure."""

from logging import DEBUG, FileHandler, Formatter, StreamHandler, getLogger
from os.path import join
from sys import stdout


class Log():
    """Log is the ajsonapi logger."""

    def __init__(self, file_path='', file_name=''):
        # pylint: disable=invalid-name
        self.logger = getLogger('ajsonapi')
        self.logger.setLevel(DEBUG)

        fm = Formatter(
            '[%(process)d][%(asctime)s.%(msecs)03d] '
            '%(levelname)s: %(message)s',
            datefmt='%Y-%m-%d %T')
        sh = StreamHandler(stdout)
        sh.setLevel(DEBUG)
        sh.setFormatter(fm)
        self.logger.addHandler(sh)

        if file_name:
            fh = FileHandler(join(file_path, file_name))
            fh.setFormatter(fm)
            self.logger.addHandler(fh)

    def debug(self, msg, *args):
        """Log a message at the debug log level."""
        self.logger.debug(msg, *args)

    def info(self, msg, *args):
        """Log a message at the info log level."""
        self.logger.info(msg, *args)

    def warning(self, msg, *args):
        """Log a message at the warning log level."""
        self.logger.warning(msg, *args)

    def error(self, msg, *args):
        """Log a message at the error log level."""
        self.logger.error(msg, *args)

    def critical(self, msg, *args):
        """Log a message at the critical log level."""
        self.logger.critical(msg, *args)


log = Log()  # pylint: disable=invalid-name



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module types provided types that can be used in the Attribute constructor.
"""


class Type:
    """Abstract base class for supported database types."""

    # pylint: disable=too-few-public-methods

    @staticmethod
    def sql(value):
        """Converts value into sql value."""
        return value


class Bool(Type):
    """Type for booleans."""

    # pylint: disable=too-few-public-methods

    name = 'BOOLEAN'


# Deprecating
Boolean = Bool


class DateTime(Type):
    """Type for date times."""

    # pylint: disable=too-few-public-methods

    name = 'TIMESTAMP'


class DateTimeTZ(Type):
    """Type for date times with time zone information."""

    # pylint: disable=too-few-public-methods

    name = 'TIMESTAMP WITH TIME ZONE'


# Deprecating
DateTimeTimeZone = DateTimeTZ


class Float32(Type):
    """Type for 32 bit floating point numbers."""

    # pylint: disable=too-few-public-methods

    name = 'FLOAT'


class Float64(Type):
    """Type for 64 bit floating point numbers."""

    # pylint: disable=too-few-public-methods

    name = 'DOUBLE PRECISION'


class Int16(Type):
    """Type for 16 bit signed integers."""

    # pylint: disable=too-few-public-methods

    name = 'SMALLINT'


class Int32(Type):
    """Type for 32 bit signed integers."""

    # pylint: disable=too-few-public-methods

    name = 'INT'


class Int64(Type):
    """Type for 64 bit signed integers."""

    # pylint: disable=too-few-public-methods

    name = 'BIGINT'


class Json(Type):
    """Type for json data."""

    # pylint: disable=too-few-public-methods

    name = 'JSONB'


class String(Type):
    """Type for strings."""

    # pylint: disable=too-few-public-methods

    name = 'TEXT'

    @staticmethod
    def sql(value):
        sql_value = value.replace("'", "''")
        return f"'{sql_value}'"


class Uuid(Type):
    """Type for uuids."""

    # pylint: disable=too-few-public-methods

    name = 'UUID'



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module collection provides class Collection."""

from abc import abstractmethod
from asyncio import Event, Lock, create_task, gather
from itertools import chain
from re import compile as re_compile

from asyncpg.exceptions import ForeignKeyViolationError, UniqueViolationError

from ajsonapi.conversions import id_number_to_name
from ajsonapi.document import (
    ResourceObjectDocument,
    parse_request_document_data,
)
from ajsonapi.environments import GetCollectionIncludedEnvironment
from ajsonapi.errors import (
    DocumentDataDuplicateIdError,
    FilterInvalidFieldError,
    IncludeInvalidPathError,
    QueryParameterUnsupportedError,
    ResourceNotFoundError,
    SortInvalidFieldError,
)
from ajsonapi.events import CreateEvent
from ajsonapi.exceptions import ErrorsException, NonexistentRelationshipIdError
from ajsonapi.logging import log
from ajsonapi.query import parse_fields, parse_page, query_has_parameter
from ajsonapi.responses import document_response, no_data_response
from ajsonapi.uri.resource import Resource

RE_FILTER = re_compile(r'filter\[([^\]]*)\]')


class Collection(Resource):
    """Class collection represents a resource collection."""

    # pylint: disable=too-many-public-methods

    by_name = {}

    def __init__(self, name, json_api):
        self.name = name
        self.table = json_api

    async def parse_document(self, request):
        """Parses the request document for a valid resource object.

        Args:
            request (Request): the request to be parsed.

        Returns:
            A ResourceObjectDocument.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource object.
        """

        data = await parse_request_document_data(request)
        document = ResourceObjectDocument(data)
        document.parse(self)
        return document

    def parse_query(self,
                    request,
                    *,
                    allow_include=False,
                    allow_fields=False,
                    allow_filter=False,
                    allow_sort=False,
                    allow_page=False):
        """Parses the query parameters relative to this collection."""

        # pylint: disable=no-self-use,too-many-arguments

        query = {}
        errors = []

        request_query = request.query
        self.parse_include(request_query, query, errors, allow=allow_include)
        parse_fields(request_query, query, errors, allow=allow_fields)
        self.parse_filter(request_query, query, errors, allow=allow_filter)
        self.parse_sort(request_query, query, errors, allow=allow_sort)
        parse_page(request_query, query, errors, allow=allow_page)

        if errors:
            raise ErrorsException(errors)
        return query

    def parse_include(self, request_query, query, errors, allow):
        """Parses the request's include query parameter."""

        if query_has_parameter(request_query, 'include'):
            if allow:
                query['includes'] = self.includes_dict(
                    request_query['include'].split(','), errors, [])
            else:
                errors.append(QueryParameterUnsupportedError('include'))
        else:
            query['includes'] = None

    def parse_filter(self, request_query, query, errors, allow):
        """Parses the request's filter query parameter."""

        if query_has_parameter(request_query, 'filter'):
            if allow:
                query['filters'] = self.filters_list(request_query, errors)
            else:
                errors.append(QueryParameterUnsupportedError('filter'))
        else:
            query['filters'] = None

    def parse_sort(self, request_query, query, errors, allow):
        """Parses the request's sort query parameter."""

        if query_has_parameter(request_query, 'sort'):
            if allow:
                query['sort'] = self.sort_list(request_query, errors)
            else:
                errors.append(QueryParameterUnsupportedError('sort'))
        else:
            query['sort'] = None

    def includes_dict(self, include_parameters, errors, path):
        """Helper function for parsing request query include parameters."""
        return_dict = {}
        for include_parameter in include_parameters:
            include, *sub_includes = include_parameter.split('.', 1)
            path.append(include)
            try:
                relationship = getattr(self.table.___, include)
            except AttributeError:
                errors.append(IncludeInvalidPathError('.'.join(path)))
            else:
                return_dict.setdefault(relationship, {}).update(
                    relationship.rtable.collection.includes_dict(
                        sub_includes, errors, path))
            path.pop()
        return return_dict

    def filters_list(self, request_query, errors):
        """Helper function for parsing request query filter parameters."""

        return_list = []
        for key, value in request_query.items():
            match = RE_FILTER.match(key)
            if match:
                field_name = match.group(1)
                try:
                    field = getattr(self.table.___, field_name)
                    condition = field.filter_condition(value)
                except AttributeError:
                    errors.append(FilterInvalidFieldError(field_name))
                    continue
                return_list.append(condition)
        return return_list

    def sort_list(self, request_query, errors):
        """Helper function for parsing request query sort parameter."""

        return_list = []
        field_names = request_query['sort'].split(',')
        for field_name in field_names:
            if field_name[0] == '-':
                field_name = field_name[1:]
                order = f'{field_name} DESC'
            else:
                order = f'{field_name} ASC'
            try:
                _ = getattr(self.table.___, field_name)
            except AttributeError:
                errors.append(SortInvalidFieldError(field_name))
                continue
            return_list.append(order)
        return return_list

    async def get(self, query):
        """Produces the response for a GET /{collection} request."""

        links = {'self': self.path()}
        return document_response(await self.to_document(query=query,
                                                        links=links))

    async def post(self, document, query):
        """Produces the response for a POST /{collection} request."""

        try:
            if document.id_number:
                return await self.post_with_id(document, query)
            return await self.post_without_id(document, query)
        except UniqueViolationError:
            raise ErrorsException([
                DocumentDataDuplicateIdError(f"/data/id/{document.data['id']}")
            ])

    async def post_with_id(self, document, query):
        """Produces the response for a POST /{collection} request where the
        document's resource object contains an id.
        """

        env = PostCollectionWithIdEnvironment(self, document.id_number,
                                              document.attributes,
                                              document.relationships, query)
        await env.run()
        event = CreateEvent(document.data)
        await event.broadcast()
        return no_data_response()

    async def post_without_id(self, document, query):
        """Produces the response for a POST /{collection} request where the
        document's resource object does not contain an id.
        """

        env = PostCollectionWithoutIdEnvironment(self, document.attributes,
                                                 document.relationships, query)
        await env.run()
        id_name = id_number_to_name(env.id_number)
        document.data['id'] = id_name
        path = f'{self.path()}/{id_name}'
        document.data['links'] = {'self': path}
        event = CreateEvent(document.data)
        await event.broadcast()
        return document_response({'data': document.data},
                                 status=201,
                                 headers={'Location': path})

    async def post_with_id_task(self, env):
        """Updates the database collection table based on the POST
        /{collection} request.
        """
        local_relationships = {
            self.table.lfkey_by_relationship_name[name]: value
            for name, value in env.relationships.items()
            if name in self.table.lfkey_by_relationship_name
        }
        lfkeys = local_relationships.keys()
        lfkey_values = [id_.number for id_ in local_relationships.values()]
        column_names = ', '.join(chain(['id'], env.attributes.keys(), lfkeys))
        length = 1 + len(env.attributes) + len(lfkeys)
        column_values = ', '.join(
            [f'${index}' for index in range(1, length + 1)])
        stmt = (f'INSERT INTO {self.table.name} ({column_names}) '
                f'VALUES ({column_values})')
        try:
            async with env.lock:
                async with env.connection.transaction():
                    log.debug(f'{stmt}: {env.id_number}, '
                              f'{env.attributes.values()}, {lfkey_values}')
                    await env.connection.execute(stmt, env.id_number,
                                                 *env.attributes.values(),
                                                 *lfkey_values)
        except ForeignKeyViolationError:
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def post_without_id_task(self, env):
        """Updates the database collection table based on the POST
        /{collection} request.
        """
        local_relationships = {
            self.table.lfkey_by_relationship_name[name]: value
            for name, value in env.relationships.items()
            if name in self.table.lfkey_by_relationship_name
        }
        lfkeys = local_relationships.keys()
        lfkey_values = [id_.number for id_ in local_relationships.values()]
        column_names = ', '.join(chain(env.attributes.keys(), lfkeys))
        length = len(env.attributes) + len(lfkeys)
        column_values = ', '.join(
            [f'${index}' for index in range(1, length + 1)])
        if column_values:
            stmt = (f'INSERT INTO {self.table.name} ({column_names}) '
                    f'VALUES ({column_values}) RETURNING id')
        else:
            stmt = (f'INSERT INTO {self.table.name} '
                    'DEFAULT VALUES RETURNING id')
        try:
            async with env.lock:
                async with env.connection.transaction():
                    log.debug(f'{stmt}: {env.attributes.values()}, '
                              f'{lfkey_values}')
                    env.id_number = await env.connection.fetchval(
                        stmt, *env.attributes.values(), *lfkey_values)
                    env.event.set()
        except ForeignKeyViolationError:
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def to_document(self, query=None, links=None):
        """Creates the document for the resource objects in this resource
        collection.

        Args:
            links (dict): value for the links field in the response.
        """

        data, included = await self.to_data_included(query)
        document = {'data': data}
        if included is not None:
            document['included'] = included
        if links:
            document['links'] = links
        return document

    async def to_data_included(self, query):
        """Creates the response document's data and included member values."""

        env = GetCollectionEnvironment(self, query)
        await env.run()
        return env.data, env.included

    async def get_collection_data(self, env):
        """Creates the response data member for attributes and local
        relationships.

        Specifically, after this call the environment's data member contains
        something like the following dictionary (from id to response data for
        that object):

        {
            1: {
                'type': 'centers',
                'id': '1',
                'attributes': {
                    'attr_int': 1,
                    'attr_str': 'one',
                },
                'relationships': {
                    'one_one_local': {
                        'data': {
                            'type': 'one_one_locals',
                            'id': UUID_11,
                        },
                        'links': {
                            'self': '/centers/1/relationships/one_one_local',
                            'related': '/centers/1/one_one_local',
                        },
                    },
                    'many_one': {
                        'data': {
                            'type': 'many_ones',
                            'id': UUID_31,
                        },
                        'links': {
                            'self': '/centers/1/relationships/many_ones',
                            'related': '/centers/1/many_ones',
                        },
                    },
                },
                'links': {
                    'self': '/centers/1',
                },
            },
            2: {
                ...,
            },
            ...
        }
        """
        fields = env.query['fields']
        if self in fields:
            columns = chain(
                ['id'],
                (str(col) for col in self.table.columns if col in fields[self]))
        else:
            columns = (str(col) for col in self.table.columns)
        filters = env.query['filters']
        sort = env.query['sort']
        if filters:
            if sort:
                stmt = (f"SELECT {', '.join(columns)} FROM {self.table.name} "
                        f"WHERE {' AND '.join(filters)} "
                        f"ORDER BY {', '.join(sort)}")
            else:
                stmt = (f"SELECT {', '.join(columns)} FROM {self.table.name} "
                        f"WHERE {' AND '.join(filters)}")
        else:
            if sort:
                stmt = (f"SELECT {', '.join(columns)} FROM {self.table.name} "
                        f"ORDER BY {', '.join(sort)}")
            else:
                stmt = f"SELECT {', '.join(columns)} FROM {self.table.name}"
        page = env.query['page']
        if page:
            limit = page['size']
            offset = page['number'] * limit
            stmt += f' LIMIT {limit} OFFSET {offset}'
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        env.data_by_lid = {
            record['id']: self.record_to_data(record) for record in records
        }
        env.event.set()

    async def get_collection_included(self, env, ids):
        """Collects part of the 'included' member for a GET
        /{collection}?include response.
        """
        fields = env.query['fields']
        if self in fields:
            columns = chain(
                ['id'],
                (str(col) for col in self.table.columns if col in fields[self]))
        else:
            columns = (str(col) for col in self.table.columns)
        stmt = (f"SELECT {', '.join(columns)} "
                f"FROM {self.table.name} WHERE id IN ({ids})")
        ### Optimize the stmt!
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        env.included_by_id = {
            record['id']: self.record_to_data(record) for record in records
        }
        env.event.set()

    async def get_object_included(self, env, ids):
        """Collects part of the 'included' member for a GET
        /{collection}?include response.
        """
        fields = env.query['fields']
        if self in fields:
            columns = chain(
                ['id'],
                (str(col) for col in self.table.columns if col in fields[self]))
        else:
            columns = (str(col) for col in self.table.columns)
        stmt = (f"SELECT {', '.join(columns)} "
                f"FROM {self.table.name} WHERE id IN ({ids})")
        ### Optimize the stmt!
        async with env.lock:
            log.debug(f'{stmt}: {env.object_id}')
            records = await env.connection.fetch(stmt, env.object_id)
        env.included_by_id = {
            record['id']: self.record_to_data(record) for record in records
        }
        env.event.set()

    def record_to_data(self, record):
        """Converts a asyncpg record into a JSON API data member."""
        attributes = {}
        relationships = {}
        for col, value in record.items():
            if col == 'id':
                id_name = id_number_to_name(value)
            elif col in self.table.attribute_names:
                attributes[col] = value
            else:  # col is a lfkey
                rel = self.table.relationship_by_lfkey[col]
                if value:
                    relationship_data = {
                        'type': rel.rtable.collection.name,
                        'id': id_number_to_name(value)
                    }
                else:
                    relationship_data = None
                relationships[rel.name] = {
                    'data': relationship_data,
                    'links': {
                        'self': (f'/{self.name}/{id_name}/'
                                 f'relationships/{rel.name}'),
                        'related': (f'/{self.name}/{id_name}/'
                                    f'{rel.name}'),
                    }
                }
        data = {
            'type': self.name,
            'id': id_name,
        }
        if attributes:
            data['attributes'] = attributes
        if relationships:
            data['relationships'] = relationships
        return data

    def path(self):
        """Creates the path to the collection."""

        return f'/{self.name}'


class GetCollectionEnvironment:
    """Class GetCollectionEnvironment is the execution environment for a
    GET /{collection} request.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, collection, query):
        self.collection = collection
        self.query = query
        self.tasks = []
        self.connection = None
        self.lock = Lock()
        self.data_environment = None
        self.data = None
        self.included_environments = []
        self.included = None

    def create_data_environment(self):
        """Creates the environment for collecting the data member of the GET
        /{collection} response.
        """
        self.data_environment = GetCollectionDataEnvironment(self)

    def create_included_environments(self, ids, includes):
        """Creates the environment for collecting the included member of the
        GET /{collection} response.
        """
        if includes is None:
            return
        for relationship, remote_includes in includes.items():
            remote_collection = relationship.rtable.collection
            remote_ids = relationship.collection_remote_ids(self.query, ids)
            self.included_environments.append(
                GetCollectionIncludedEnvironment(remote_collection, remote_ids,
                                                 self))
            if remote_includes:
                self.create_included_environments(remote_ids, remote_includes)

    def create_tasks(self):
        """Create all tasks for querying the database to create a GET
        /{collection} response.
        """

        self.create_data_environment()
        self.create_included_environments(None, self.query['includes'])
        self.tasks = [
            create_task(env.run()) for env in chain([self.data_environment],
                                                    self.included_environments)
        ]

    async def run(self):
        """Executes the GET /{collection} tasks."""

        async with self.collection.table.pool.acquire() as self.connection:
            async with self.connection.transaction():
                self.create_tasks()
                await gather(*self.tasks)
        self.data = list(self.data_environment.data_by_lid.values())
        if self.query['includes']:
            self.included = list(
                chain.from_iterable(env.included_by_id.values()
                                    for env in self.included_environments))


class GetCollectionDataEnvironment:
    """Class GetCollectionDataEnvironment is the execution environment for the
    data part of the GET /{collection} response.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, env):
        self.collection = env.collection
        self.query = env.query
        self.tasks = []
        self.connection = env.connection
        self.lock = env.lock
        self.data_by_lid = {}
        self.event = Event()

    def create_tasks(self):
        """Create all tasks for querying the database to create the data part
        of a GET /{collection} response.
        """
        self.tasks = [create_task(self.collection.get_collection_data(self))]
        query_fields = self.query['fields']
        if self.collection in query_fields:
            self.tasks.extend([
                create_task(rel.get_collection_data(self))
                for rel in query_fields[self.collection]
                if rel in self.collection.table.remote_relationships
            ])
        else:
            self.tasks.extend([
                create_task(rel.get_collection_data(self))
                for rel in self.collection.table.remote_relationships
            ])

    async def run(self):
        """Executes the GET /{collection} data tasks."""
        self.create_tasks()
        await gather(*self.tasks)


class PostCollectionEnvironment:
    """Class PostCollectionEnvironment is the execution environment for a POST
    /{collection} request.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, collection, id_number, attributes, relationships, query):
        # pylint: disable=too-many-arguments

        self.collection = collection
        self.id_number = id_number
        self.attributes = attributes
        self.relationships = relationships
        self.query = query
        self.connection = None
        self.lock = Lock()
        self.tasks = []
        self.gather_future = None
        self.event = Event()  # To indicate that self.id_number exists

    @abstractmethod
    def create_tasks(self):
        """Create all tasks for updating the database in response to a POST
        /{collection} request.
        """

    async def run(self):
        """Executes the POST /{collection} tasks."""
        async with self.collection.table.pool.acquire() as self.connection:
            async with self.connection.transaction():
                try:
                    async with self.connection.transaction():
                        self.create_tasks()
                        self.gather_future = gather(*self.tasks)
                        await self.gather_future
                except NonexistentRelationshipIdError:
                    env = VerifyRelationshipIdsEnvironment(
                        self.collection.table, self.relationships,
                        self.connection)
                    await env.run()

    def cancel(self):
        """Cancels the environment."""

        self.gather_future.cancel()


class PostCollectionWithIdEnvironment(PostCollectionEnvironment):
    """Class PostCollectionWithIdEnvironment is the execution environment for
    a POST /{collection} request where the requests document's data contains
    an id.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, collection, id_number, attributes, relationships, query):
        # pylint: disable=too-many-arguments

        super().__init__(collection, id_number, attributes, relationships,
                         query)
        self.event.set()

    def create_tasks(self):
        """Create all tasks for updating the database in response to a POST
        /{collection} request where the request document's data contains an
        id.
        """
        self.tasks = [create_task(self.collection.post_with_id_task(self))]
        self.tasks.extend([
            create_task(rel.post_collection_task(self))
            for rel in self.collection.table.remote_relationships
            if rel.name in self.relationships
        ])


class PostCollectionWithoutIdEnvironment(PostCollectionEnvironment):
    """Class PostCollectionWithoutIdEnvironment is the execution environment
    for a POST /{collection} request where the request document's data does
    not contain an id.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, collection, attributes, relationships, query):
        # pylint: disable=too-many-arguments
        super().__init__(collection, None, attributes, relationships, query)

    def create_tasks(self):
        """Create all tasks for updating the database in response to a POST
        /{collection} request where the request document's data does not
        contain an id.
        """
        self.tasks = [create_task(self.collection.post_without_id_task(self))]
        self.tasks.extend([
            create_task(rel.post_collection_task(self))
            for rel in self.collection.table.remote_relationships
            if rel.name in self.relationships
        ])


class VerifyRelationshipIdsEnvironment:
    """Class VerifyRelationshipIdsEnvironment is the execution environment for
    verifying the existence of relationship ids.
    """

    def __init__(self, table, relationships, connection):
        self.table = table
        self.ids_by_relationship = {
            getattr(table.___, name): ids
            for name, ids in relationships.items()
        }
        self.connection = connection
        self.lock = Lock()
        self.tasks = []
        self.errors = []

    def create_tasks(self):
        """Create all tasks for verifying the document's data/id members for
        existence.
        """
        self.tasks = [
            create_task(relationship.verify_data_ids_exist(self, ids))
            for relationship, ids in self.ids_by_relationship.items()
        ]

    async def run(self):
        """Raises ErrorsException with errors for all the document's
        nonexistent relationship ids.
        """
        async with self.connection.transaction():
            self.create_tasks()
            await gather(*self.tasks)
            raise ErrorsException(self.errors)


def parse_collection(request):
    """Gets a collection associated with a request.

    Args:
        request: Incoming Http(s) request.

    Exceptions:
        ErrorsException: Exception containing a 'resource not found' error.
    """

    collection_name = request.match_info['collection']
    try:
        return Collection.by_name[collection_name]
    except KeyError:
        raise ErrorsException([ResourceNotFoundError(f'/{collection_name}')])



# Copyright © 2019-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module events provides events for changes to resource objects in the object
model."""

from abc import abstractmethod


class Event:
    """Base class for events"""

    # pylint: disable=too-few-public-methods

    app = None

    @classmethod
    def connect(cls, app):
        """Connects events to the app."""
        cls.app = app

    def __init__(self, data):
        self.data = data

    @abstractmethod
    def stream_event(self):
        """Converts the journal event into a stream event."""

    async def publish(self):
        """Publishes the event on the event jounal."""

    async def broadcast(self):
        """Tells the application to broadcast the event."""
        await self.app.broadcast(self)


class CreateEvent(Event):
    """Event to signal the creation of a resource object"""

    # pylint: disable=too-few-public-methods

    def stream_event(self):
        return {
            'op': 'create',
            'data': self.data,
        }


class UpdateEvent(Event):
    """Event to signal a change of a resource object"""

    # pylint: disable=too-few-public-methods

    def stream_event(self):
        return {
            'op': 'update',
            'data': self.data,
        }


class DeleteEvent(Event):
    """Event to signal the deletion of a resource object"""

    # pylint: disable=too-few-public-methods

    def stream_event(self):
        return {
            'op': 'delete',
            'data': self.data,
        }



# Copyright © 2019-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module uri.related provides things related to related resource URIs."""

from abc import abstractmethod
from asyncio import Event, Lock, create_task, gather
from itertools import chain

from ajsonapi.conversions import id_number_to_name
from ajsonapi.environments import GetObjectIncludedEnvironment
from ajsonapi.errors import (
    ResourceNotFoundError,
    ResourceNotFoundMalformedIdError,
)
from ajsonapi.exceptions import ErrorsException
from ajsonapi.id_value import IdValue
from ajsonapi.logging import log
from ajsonapi.responses import document_response
from ajsonapi.uri.collection import Collection
from ajsonapi.uri.resource import Resource


class RelatedResource(Resource):
    """Class RelatedResource represents the related resource URI
    (/{collection}/{id}/{related_resource}).
    """

    def __init__(self, name, collection, id_, relationship):
        self.related_resource_name = name
        self.collection = collection
        self.id = id_  # pylint: disable=invalid-name
        self.relationship = relationship
        self.table = collection.table

    async def get(self, query):
        """Produces the response for the GET
        /{collection}/{id}/{related_resource} request.
        """
        links = {'self': self.path()}
        return document_response(await self.to_document(query=query,
                                                        links=links))

    async def to_document(self, query=None, links=None):
        """Creates the document for the resource object(s) that are part of
        the related resource.
        """

        data, included = await self.to_data_included(query)
        document = {'data': data}
        if included is not None:
            document['included'] = included
        if links:
            document['links'] = links
        return document

    @abstractmethod
    def is_to_many(self):
        """Returns if the related resource is a to-many related resource."""

    @abstractmethod
    async def to_data_included(self, query):
        """Creates the document's data and included member values for this
        related resource.
        """

    @abstractmethod
    async def get_related_data(self, env):
        """Creates the response data member for attributes and local related
        resources.

        Specifically, after this call the environment's data or data_by_lid
        member is updated."""

    async def verify_id_exists(self, env):
        """Raises an exception in case the id in the URL does not exist."""
        stmt = f'SELECT FROM {self.table.name} WHERE id = $1'
        record = await env.connection.fetchrow(stmt, self.id.number)
        if record is None:
            env.cancel()
            raise ErrorsException([
                ResourceNotFoundError(f'/{self.collection.name}/{self.id.name}')
            ])

    def path(self):
        """Creates the path to the relationship."""
        return (f'/{self.collection.name}/{self.id.name}/'
                f'{self.related_resource_name}')


class ToOneRelatedResource(RelatedResource):
    """Class ToOneRelatedResource represents the to-one related resource URI
    (/{collection}/{id}/{related_resource}).
    """

    def is_to_many(self):
        return False

    async def to_data_included(self, query):
        """Creates the document's data and included member values for this
        related resources.
        """

        env = GetToOneRelatedResourceEnvironment(self, query)
        await env.run()
        return env.data, env.included

    def record_to_data(self, record):
        """Converts a asyncpg record into a JSON API data member."""
        attributes = {}
        relationships = {}
        for col, value in record.items():
            if col == 'id':
                id_name = id_number_to_name(value)
            elif col in self.relationship.rtable.attribute_names:
                attributes[col] = value
            else:  # col is a lfkey
                rel = self.relationship.rtable.relationship_by_lfkey[col]
                relationships[rel.name] = self.to_relationships(rel, value)
        data = {
            'type': self.relationship.rtable.collection.name,
            'id': id_name,
        }
        if attributes:
            data['attributes'] = attributes
        if relationships:
            data['relationships'] = relationships
        return data

    @abstractmethod
    async def get_related_data(self, env):
        pass


class OneToOneLocalRelatedResource(ToOneRelatedResource):
    """Class OneToOneLocalRelatedResource represents the URI
    (/{collection}/{id}/{related_resource}) for a one-to-one related resource
    with a local foreign id.
    """

    async def get_related_data(self, env):
        query_fields = env.query['fields']
        rtable = self.relationship.rtable
        if rtable.collection in query_fields:
            columns = chain(['id'], (str(col)
                                     for col in rtable.columns
                                     if col in query_fields[rtable.collection]))
        else:
            columns = (str(col) for col in rtable.columns)
        stmt = (f'SELECT {", ".join(columns)} '
                f'FROM {rtable.name} '
                f'WHERE id = ('
                f' SELECT {self.relationship.lfkey}'
                f' FROM {self.table.name}'
                f' WHERE id = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {self.id.number}')
            record = await env.connection.fetchrow(stmt, self.id.number)
            if record is None:
                await self.verify_id_exists(env)
                env.data = None
                env.event.set()
                return
        env.data = self.record_to_data(record)
        env.event.set()


class OneToOneRemoteRelatedResource(ToOneRelatedResource):
    """Class OneToOneRemoteRelatedResource represents the URI
    (/{collection}/{id}/{related_resource}) for a one-to-one related resource
    with a remote foreign id.
    """

    async def get_related_data(self, env):
        query_fields = env.query['fields']
        rtable = self.relationship.rtable
        if rtable.collection in query_fields:
            columns = chain(['id'], (str(col)
                                     for col in rtable.columns
                                     if col in query_fields[rtable.collection]))
        else:
            columns = (str(col) for col in rtable.columns)
        # We could remove rfkey because we know it already.
        stmt = (f'SELECT {", ".join(columns)} '
                f'FROM {self.relationship.rtable.name} '
                f'WHERE {self.relationship.rfkey} = $1')
        async with env.lock:
            log.debug(f'{stmt}: {self.id.number}')
            record = await env.connection.fetchrow(stmt, self.id.number)
            if record is None:
                await self.verify_id_exists(env)
                env.data = None
                env.event.set()
                return
        env.data = self.record_to_data(record)
        env.event.set()


class ManyToOneRelatedResource(ToOneRelatedResource):
    """Class ManyToOneRelatedResource represents the URI
    (/{collection}/{id}/{related_resource}) for a many-to-one related
    resource.
    """

    async def get_related_data(self, env):
        query_fields = env.query['fields']
        rtable = self.relationship.rtable
        if rtable.collection in query_fields:
            columns = chain(['id'], (str(col)
                                     for col in rtable.columns
                                     if col in query_fields[rtable.collection]))
        else:
            columns = (str(col) for col in self.relationship.rtable.columns)
        stmt = (f'SELECT {", ".join(columns)} '
                f'FROM {self.relationship.rtable.name} '
                f'WHERE id = ('
                f' SELECT {self.relationship.lfkey}'
                f' FROM {self.table.name}'
                f' WHERE id = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {self.id.number}')
            record = await env.connection.fetchrow(stmt, self.id.number)
            if record is None:
                await self.verify_id_exists(env)
                env.data = None
                env.event.set()
                return
        env.data = self.record_to_data(record)
        env.event.set()


class ToManyRelatedResource(RelatedResource):
    """Class ToManyRelatedResource represents the to-one related resource URI
    (/{collection}/{id}/{related_resource}).
    """

    def is_to_many(self):
        return True

    async def to_data_included(self, query):
        """Creates the document's data and included member values for this
        related resources.
        """

        env = GetToManyRelatedResourceEnvironment(self, query)
        await env.run()
        return env.data, env.included

    def record_to_data(self, record):
        """Converts a asyncpg record into a JSON API data member."""
        attributes = {}
        relationships = {}
        for col, value in record.items():
            if col == 'id':
                id_name = id_number_to_name(value)
            elif col in self.relationship.rtable.attribute_names:
                attributes[col] = value
            else:  # col is a lfkey
                rel = self.relationship.rtable.relationship_by_lfkey[col]
                if value:
                    relationship_data = {
                        'type': rel.rtable.collection.name,
                        'id': id_number_to_name(value)
                    }
                else:
                    relationship_data = None
                relationships[rel.name] = {
                    'data': relationship_data,
                    'links': {
                        'self': (f'/{self.relationship.name}/{id_name}/'
                                 f'relationships/{rel.name}'),
                        'related': (f'/{self.relationship.name}/{id_name}/'
                                    f'{rel.name}'),
                    }
                }
        data = {
            'type': self.relationship.rtable.collection.name,
            'id': id_name,
        }
        if attributes:
            data['attributes'] = attributes
        if relationships:
            data['relationships'] = relationships
        return data

    @abstractmethod
    async def get_related_data(self, env):
        pass


class OneToManyRelatedResource(ToManyRelatedResource):
    """Class OneToManyRelatedResource represents the URI
    (/{collection}/{id}/{related_resource}) for a one-to-many related
    resource.
    """

    async def get_related_data(self, env):
        query_fields = env.query['fields']
        rtable = self.relationship.rtable
        if rtable.collection in query_fields:
            columns = chain(['id'], (str(col)
                                     for col in rtable.columns
                                     if col in query_fields[rtable.collection]))
        else:
            columns = (str(col) for col in rtable.columns)
        # We could remove rfkey because we know it already
        filters = env.query['filters']
        sort = env.query['sort']
        if filters:
            if sort:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE {self.relationship.rfkey} = $1 AND '
                        f'{" AND ".join(filters)} '
                        f'ORDER BY {", ".join(sort)}')
            else:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE {self.relationship.rfkey} = $1 AND '
                        f'{" AND ".join(filters)}')
        else:
            if sort:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE {self.relationship.rfkey} = $1 '
                        f'ORDER BY {", ".join(sort)}')
            else:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE {self.relationship.rfkey} = $1')
        async with env.lock:
            log.debug(f'{stmt}: {self.id.number}')
            records = await env.connection.fetch(stmt, self.id.number)
            if records == []:
                await self.verify_id_exists(env)
        env.data_by_lid = {
            record['id']: self.record_to_data(record) for record in records
        }
        env.event.set()


class ManyToManyRelatedResource(ToManyRelatedResource):
    """Class ManyToManyRelatedResource represents the URI
    (/{collection}/{id}/{related_resource}) for a many-to-many related
    resource.
    """

    async def get_related_data(self, env):
        query_fields = env.query['fields']
        rtable = self.relationship.rtable
        if rtable.collection in query_fields:
            columns = chain(['id'], (str(col)
                                     for col in rtable.columns
                                     if col in query_fields[rtable.collection]))
        else:
            columns = (str(col) for col in rtable.columns)
        filters = env.query['filters']
        sort = env.query['sort']
        if filters:
            if sort:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE id IN ('
                        f' SELECT {self.relationship.rafkey}'
                        f' FROM {self.relationship.atable.name}'
                        f' WHERE {self.relationship.lafkey} = $1) AND'
                        f' {" AND ".join(filters)} '
                        f'ORDER BY {", ".join(sort)}')
            else:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE id IN ('
                        f' SELECT {self.relationship.rafkey}'
                        f' FROM {self.relationship.atable.name}'
                        f' WHERE {self.relationship.lafkey} = $1) AND'
                        f' {" AND ".join(filters)}')
        else:
            if sort:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE id IN ('
                        f' SELECT {self.relationship.rafkey}'
                        f' FROM {self.relationship.atable.name}'
                        f' WHERE {self.relationship.lafkey} = $1) '
                        f'ORDER BY {", ".join(sort)}')
            else:
                stmt = (f'SELECT {", ".join(columns)} '
                        f'FROM {self.relationship.rtable.name} '
                        f'WHERE id IN ('
                        f' SELECT {self.relationship.rafkey}'
                        f' FROM {self.relationship.atable.name}'
                        f' WHERE {self.relationship.lafkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {self.id.number}')
            records = await env.connection.fetch(stmt, self.id.number)
            if records == []:
                await self.verify_id_exists(env)
        env.data_by_lid = {
            record['id']: self.record_to_data(record) for record in records
        }
        env.event.set()


class GetRelatedResourceEnvironment:
    """Class GetRelatedResourceEnvironment is the execution environment
    for a GET /{collection}/{id}/{related_resource} request.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, related, query):
        self.related = related
        self.query = query
        self.tasks = []
        self.connection = None
        self.lock = Lock()
        self.data_environment = None
        self.data = None
        self.included_environments = []
        self.included = None

    @abstractmethod
    def create_data_environment(self):
        """Creates the environment for collecting the data member of a GET
        /{collection}/{id}/{related_resource} response.
        """

    def create_included_environments(self, ids, includes):
        """Creates the environment for collecting the included member of the
        GET /{collection}/{id}/{related_resource} response.
        """
        if includes is None:
            return
        for relationship, remote_includes in includes.items():
            remote_collection = relationship.rtable.collection
            remote_ids = relationship.object_remote_ids(ids)
            self.included_environments.append(
                GetObjectIncludedEnvironment(self.related.id.number,
                                             remote_collection, remote_ids,
                                             self))
            if remote_includes:
                self.create_included_environments(remote_ids, remote_includes)

    def create_tasks(self):
        """Create all tasks for querying the database to create a GET
        /{collection}/{id}/{related_resource} response.
        """
        self.create_data_environment()
        relationship = self.related.relationship
        self.create_included_environments(relationship.object_remote_ids(),
                                          self.query['includes'])
        self.tasks = [
            create_task(env.run()) for env in chain([self.data_environment],
                                                    self.included_environments)
        ]

    async def run(self):
        """Executes the GET /{collection}/{id}/{related_resource} tasks."""

        async with self.related.table.pool.acquire() as self.connection:
            async with self.connection.transaction():
                self.create_tasks()
                await gather(*self.tasks)
        self.data = self.data_environment.get_data()
        if self.query['includes']:
            self.included = list(
                chain.from_iterable(env.included_by_id.values()
                                    for env in self.included_environments))


class GetToOneRelatedResourceEnvironment(GetRelatedResourceEnvironment):
    """Class GetToOneRelatedResourceEnvironment is the execution environment
    for a GET /{collection}/{id}/{related_resource} request where
    related_resource corresponds to a to-one relationship.
    """

    def create_data_environment(self):
        self.data_environment = GetToOneRelatedResourceDataEnvironment(self)


class GetToManyRelatedResourceEnvironment(GetRelatedResourceEnvironment):
    """Class GetToManyRelatedResourceEnvironment is the execution environment
    for a GET /{collection}/{id}/{related_resource} request where
    related_resource corresponds to a to-many relationship.
    """

    def create_data_environment(self):
        self.data_environment = GetToManyRelatedResourceDataEnvironment(self)


class GetRelatedResourceDataEnvironment:
    """Class GetRelatedResourceDataEnvironment is the execution environment
    for a GET /{collection}/{id}/{related_resource} request.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, env):
        self.related = env.related
        self.query = env.query
        self.tasks = []
        self.gather_future = None
        self.connection = env.connection
        self.lock = env.lock
        self.event = Event()

    @abstractmethod
    def create_tasks(self):
        """Create all tasks for querying the database to create the data
        member of a GET /{collection}/{id}/{related_resource} response.
        """

    async def run(self):
        """Executes the GET /{collection}/{id}/{related_resource} tasks."""

        self.create_tasks()
        self.gather_future = gather(*self.tasks)
        await self.gather_future

    def cancel(self):
        """Cancels the environment."""

        self.gather_future.cancel()


class GetToOneRelatedResourceDataEnvironment(GetRelatedResourceDataEnvironment):
    """Class GetToOneRelatedResourceDataEnvironment is the execution
    environment for the data part of a GET
    /{collection}/{id}/{related_resource} response where related_resource
    corresponds to a to-one relationship.
    """

    def __init__(self, env):
        super().__init__(env)
        self.data = None

    def create_tasks(self):
        """Create all tasks for querying the database to create the data
        member of a GET /{collection}/{id}/{related_resource} response where
        related_resource corresponds to a to-one relationship.
        """
        self.tasks = [create_task(self.related.get_related_data(self))]
        query_fields = self.query['fields']
        if self.related.relationship.rtable.collection in query_fields:
            self.tasks.extend([
                create_task(rel.get_to_one_related_data(self))
                for rel in query_fields[
                    self.related.relationship.rtable.collection]
                if rel in self.related.relationship.rtable.remote_relationships
            ])
        else:
            self.tasks.extend([
                create_task(rel.get_to_one_related_data(self))
                for rel in self.related.relationship.rtable.remote_relationships
            ])

    def get_data(self):
        """Returns the data collected by executing the environment."""
        return self.data


class GetToManyRelatedResourceDataEnvironment(GetRelatedResourceDataEnvironment
                                             ):
    """Class GetToManyRelatedResourceDataEnvironment is the execution
    environment for the data part of a GET
    /{collection}/{id}/{related_resource} response where related_resource
    corresponds to a to-many relationship.
    """

    def __init__(self, env):
        super().__init__(env)
        self.data_by_lid = {}

    def create_tasks(self):
        """Create all tasks for querying the database to create the data
        member of a GET /{collection}/{id}/{related_resource} response where
        related_resource corresponds to a to-many relationship.
        """
        self.tasks = [create_task(self.related.get_related_data(self))]
        query_fields = self.query['fields']
        if self.related.relationship.rtable.collection in query_fields:
            self.tasks.extend([
                create_task(rel.get_to_many_related_data(self))
                for rel in query_fields[
                    self.related.relationship.rtable.collection]
                if rel in self.related.relationship.rtable.remote_relationships
            ])
        else:
            self.tasks.extend([
                create_task(rel.get_to_many_related_data(self))
                for rel in self.related.relationship.rtable.remote_relationships
            ])

    def get_data(self):
        """Returns the data collected by executing the environment."""
        return list(self.data_by_lid.values())


def make_related_resource(collection_name, id_name, related_resource_name):
    """Returns the related resource URI
    (/{collection}/{id}/{related_resource}) corresponding to the user provided
    model relationships.

    Args:
        collection_name (str): Name of the collection.
        id_name (str): Name of the id.
        related_resource_name (str): Name of the related resource.

    Returns:
        An instantiated object of the correct derived class of RelatedResource.
    """

    try:
        collection = Collection.by_name[collection_name]
    except KeyError:
        raise ErrorsException([ResourceNotFoundError(f'/{collection_name}')])

    try:
        id_ = IdValue(id_name)
    except ValueError:
        raise ErrorsException(
            [ResourceNotFoundMalformedIdError(f'/{collection_name}/{id_name}')])

    try:
        relationship = getattr(collection.table.___, related_resource_name)
    except AttributeError:
        raise ErrorsException([
            ResourceNotFoundError(f'/{collection_name}/{id_name}/'
                                  f'{related_resource_name}')
        ])

    return relationship.make_uri_related_resource(related_resource_name,
                                                  collection, id_)


def parse(request):
    """Gets a related resource associated with a request.

    Args:
        request: Incoming Http(s) request.

    Returns:
        A URI related resource.

    Exceptions:
        ErrorsException: Exception containing a 'resource not found' error.
    """

    collection_name = request.match_info['collection']
    id_name = request.match_info['id']
    related_resource_name = request.match_info['related_resource']
    return make_related_resource(collection_name, id_name,
                                 related_resource_name)



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module document deals with anything related to request documents."""

from json import JSONDecodeError

from ajsonapi.conversions import id_name_to_number, uuid_name_to_number
from ajsonapi.errors import (
    DocumentDataInvalidAttributeError,
    DocumentDataInvalidIdError,
    DocumentDataInvalidRelationshipError,
    DocumentDataInvalidTypeError,
    DocumentDataMalformedError,
    DocumentDataMalformedIdError,
    DocumentDataMalformedRelationshipError,
    DocumentDataMalformedUuidError,
    DocumentDataMissingError,
    DocumentDataMissingIdError,
    DocumentDataMissingTypeError,
    DocumentNotJsonError,
)
from ajsonapi.exceptions import ErrorsException, TypeNameError
from ajsonapi.id_value import IdValue


class Document:
    """Representation of a request document."""

    # pylint: disable=too-few-public-methods

    def __init__(self, data):
        self.data = data

    def parse_type(self, expected_type_name, errors):
        """Verifies that the document's resource object contains the correct
        type.

        Args:
            expected_type_name (str): expected type of the resource (identifier)
                object in the document.
            errors (list): list of errors found so far verifying the document
                describes a valid resourse (identifier) object.
        """

        try:
            type_name = self.data['type']
            if type_name != expected_type_name:
                errors.append(
                    DocumentDataInvalidTypeError(f'/data/type/{type_name}'))
        except TypeError:
            raise ErrorsException([DocumentDataMalformedError()])
        except KeyError:
            errors.append(DocumentDataMissingTypeError('/data'))


class ResourceObjectDocument(Document):
    """Representation of a request document with a resource object."""

    def __init__(self, data):
        super().__init__(data)
        self.id_number = None
        self.attributes = None
        self.relationships = None

    def parse(self, collection, id_name=None):
        """Parses the document for a valid resource object.

        Args:
            collection (Collection): collection from which the resource object
                should be a member.
            id_name (str): valid id name for the resource object.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource object.
        """

        errors = []
        self.parse_type(collection.name, errors)
        self.parse_id(id_name, errors)
        self.parse_attributes(collection, errors)
        self.parse_relationships(collection, errors)
        if errors:
            raise ErrorsException(errors)

    def parse_id(self, expected_id_name, errors):
        """Parses the document's resource object for a valid id.

        Args:
            expected_id_name (str): expected id of the resource object in the
                document.
            errors (list): list of errors found so far verifying the document
                describes a valid resourse object.
        """

        if 'id' in self.data:
            id_name = self.data['id']
            try:
                if expected_id_name:
                    if id_name != expected_id_name:
                        errors.append(
                            DocumentDataInvalidIdError(
                                'Invalid resource object.',
                                f'/data/id/{id_name}'))
                    self.id_number = id_name_to_number(id_name)
                else:
                    self.id_number = uuid_name_to_number(id_name)
            except ValueError:
                errors.append(
                    DocumentDataMalformedUuidError(f'/data/id/{id_name}'))
        else:
            if expected_id_name:
                errors.append(DocumentDataMissingIdError())
            self.id_number = None

    def parse_attributes(self, collection, errors):
        """Parses the document's resource object for valid attributes.

        Args:
            collection (Collection): collection of which the resource object
                should be a member.
            errors (list): list of errors found so far verifying the document
                describes a valid resourse object.
        """

        self.attributes = self.data.get('attributes', {})
        invalid_attribute_names = (
            set(self.attributes.keys()) -
            {attr.name for attr in collection.table.attributes})
        if invalid_attribute_names:
            errors.extend([
                DocumentDataInvalidAttributeError(
                    f'/data/attributes/{attribute_name}')
                for attribute_name in invalid_attribute_names
            ])

    def parse_relationships(self, collection, errors):
        """Parses the document's resource object for valid relationships.

        Args:
            collection (Collection): collection from which the resource object
                should be a member.
            errors (list): list of errors found so far verifying the document
                describes a valid resourse object.
        """

        data_relationships = self.data.get('relationships', {})
        invalid_data_relationship_names = (
            set(data_relationships.keys()) -
            {rel.name for rel in collection.table.relationships})
        if invalid_data_relationship_names:
            errors.extend([
                DocumentDataInvalidRelationshipError(
                    f'/data/relationships/{data_relationship_name}')
                for data_relationship_name in invalid_data_relationship_names
            ])
        self.relationships = {}
        for relationship_name, value in data_relationships.items():
            if relationship_name in invalid_data_relationship_names:
                continue
            relationship = getattr(collection.table.___, relationship_name)
            try:
                relationship_data = value['data']
            except TypeError:
                errors.append(
                    DocumentDataMalformedRelationshipError(
                        f'/data/relationships/{relationship_name}'))
                continue
            data_rio_ids = relationship.verify_data_rios(
                relationship_name, relationship_data, errors)
            self.relationships[relationship_name] = data_rio_ids


class ResourceIdentifierObjectDocument(Document):
    """Representation of a request document with a resource identification
    object.
    """

    def __init__(self, data):
        super().__init__(data)
        self.id = None  # pylint: disable=invalid-name

    def parse(self, expected_type_name):
        """Parses the document for a valid resource identifier object.

        Args:
            expected_type_name (str): expected type name of the resource
                identifier object.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource identifier object.
        """

        errors = []
        if self.data is not None:
            self.parse_type(expected_type_name, errors)
            self.parse_id(errors)
            if errors:
                raise ErrorsException(errors)

    def parse_id(self, errors):
        """Verifies that the document's resource identifier object contains a
        valid id.

        Args:
            errors (list): list of errors found so far verifying the document
                describes a valid resourse object.
        """
        try:
            id_name = self.data['id']
            try:
                self.id = IdValue(id_name)
            except ValueError:
                errors.append(
                    DocumentDataMalformedIdError(
                        'Invalid resource identifier object.',
                        f'/data/id/{id_name}'))
        except KeyError:
            errors.append(DocumentDataMissingIdError('/data'))


def parse_roi_type(roi, index, expected_type_name, errors):
    """Parses the resource object indetifier for an invalid type."""

    try:
        roi_type_name = roi['type']
        if roi_type_name != expected_type_name:
            errors.append(
                DocumentDataInvalidTypeError(
                    f'/data/{index}/type/{roi_type_name}'))
    except KeyError:
        errors.append(DocumentDataMissingTypeError(f'/data/{index}'))


def parse_roi_id(roi, index, errors):
    """Parses the resource object indentier for an invalid id."""

    try:
        roi_id_name = roi['id']
        try:
            id_name_to_number(roi_id_name)
        except ValueError:
            errors.append(
                DocumentDataMalformedIdError(
                    'Invalid resource identifier object.',
                    f'/data/{index}/id/{roi_id_name}'))
    except KeyError:
        errors.append(DocumentDataMissingIdError(f'/data/{index}'))


class ResourceIdentifierObjectsDocument(Document):
    """Representation of a request document with a list of resource
    identification objects.
    """

    def __init__(self, data):
        super().__init__(data)
        self.ids = None

    def parse(self, expected_type_name):
        """Parses the document for a valid resource identifier objects.

        Args:
            expected_type_name (str): expected type name of the resource
                identifier object.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource identifier object.
        """

        if not isinstance(self.data, list):
            raise ErrorsException([DocumentDataMalformedError()])

        try:
            obj_type_names = {obj['type'] for obj in self.data}
            if not obj_type_names <= {expected_type_name}:
                raise TypeNameError()
            self.ids = [IdValue(obj['id']) for obj in self.data]
        except (KeyError, ValueError, TypeNameError):
            errors = []
            for index, roi in enumerate(self.data):
                parse_roi_type(roi, index, expected_type_name, errors)
                parse_roi_id(roi, index, errors)
            raise ErrorsException(errors)


async def parse_request_document_data(request):
    """Parses the request document for a valid data member.

    Args:
        request (Request): the request to be parsed.

    Returns:
        The data member of the request document.

    Raises:
        ErrorsException with an error in case the request document is
        malformed.
    """

    try:
        document = await request.json()
    except JSONDecodeError:
        raise ErrorsException([DocumentNotJsonError()])
    try:
        data = document['data']
    except KeyError:
        raise ErrorsException([DocumentDataMissingError()])
    return data



# Copyright © 2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This is a copy of https://pypi.org/project/toposort/ (thank you) without the
# use of "sorted" in the CircularDependencyError which caused trouble.
"""Module toposort provides topological sort algorithms."""

from functools import reduce as _reduce


class CircularDependencyError(ValueError):
    """CircularDependencyError indicates a circular dependency in the data
    provided to the toposort function.
    """

    def __init__(self, data):
        err = 'Circular dependencies exist among these items: {{{}}}'.format(
            ', '.join(
                '{!r}:{!r}'.format(key, value) for key, value in data.items()))
        super(CircularDependencyError, self).__init__(err)
        self.data = data


def toposort(data):
    """Dependencies are expressed as a dictionary whose keys are items
and whose values are a set of dependent items. Output is a list of
sets in topological order. The first set consists of items with no
dependences, each subsequent set consists of items that depend upon
items in the preceeding sets.
"""

    # Special case empty input.
    if len(data) == 0:
        return

    # Copy the input so as to leave it unmodified.
    data = data.copy()

    # Ignore self dependencies.
    for key, value in data.items():
        value.discard(key)
    # Find all items that don't depend on anything.
    extra_items_in_deps = _reduce(set.union, data.values()) - set(data.keys())
    # Add empty dependences where needed.
    data.update({item: set() for item in extra_items_in_deps})
    while True:
        ordered = set(item for item, dep in data.items() if len(dep) == 0)
        if not ordered:
            break
        yield ordered
        data = {
            item: (dep - ordered)
            for item, dep in data.items()
            if item not in ordered
        }
    if len(data) != 0:
        raise CircularDependencyError(data)



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Build script for setuptools"""

import setuptools

from ajsonapi import __version__

with open("README.md", "r") as fh:
    # pylint: disable=invalid-name
    long_description = fh.read()

setuptools.setup(
    name="ajsonapi",
    version=__version__,
    author="Roel van der Goot",
    author_email="roel@simplecuriosities.com",
    description="Asynchronous JSON API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/rvdg/ajsonapi",
    packages=setuptools.find_packages(),
    install_requires=['asyncio', 'aiohttp', 'asyncpg'],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module jsonapi provides class JSON_API."""

from ajsonapi.id import Id
from ajsonapi.table import Table
from ajsonapi.types import Int64


class JSON_API(Table):
    """JSON_API is the base class from which users derive object model
    classes. Classes that derive from JSON_API will result in URLs and methods
    in the web server through which to manipulate the object model.
    """
    # pylint: disable=invalid-name, too-few-public-methods

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        if not hasattr(cls, 'id'):
            cls.id = Id(Int64)
        cls.id.table_name = cls.name
        cls.columns.insert(0, cls.id)
        cls.constraints.insert(0, cls.id)


def init(pool):
    """Initializes all JSON_API subclasses."""

    for json_api in JSON_API.__subclasses__():
        json_api.pool = pool



# Copyright © 2019-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module uri.resrouce provides the Resource base class."""

from abc import ABC

from ajsonapi.conversions import id_number_to_name


class Resource(ABC):
    """Class Resource is the base for all uri resource objects."""

    # pylint: disable=too-few-public-methods

    def to_relationships(self, rel, value):
        """Converts a relationship and its record value into a JSON API
        relationships member.
        """

        # pylint: disable=no-member
        if value is None:
            relationship_data = None
        else:
            relationship_data = {
                'type': rel.rtable.collection.name,
                'id': id_number_to_name(value)
            }
        return {
            'data': relationship_data,
            'links': {
                'self': (f'/{self.collection.name}/{self.id.name}/'
                         f'relationships/{rel.name}'),
                'related': (f'/{self.collection.name}/{self.id.name}/'
                            f'{rel.name}'),
            }
        }



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module object provides class Object."""

from asyncio import Event, Lock, create_task, gather
from itertools import chain

from asyncpg.exceptions import ForeignKeyViolationError

from ajsonapi.document import ResourceObjectDocument
from ajsonapi.environments import GetObjectIncludedEnvironment
from ajsonapi.errors import (
    DeleteObjectRemotelyRelatedError,
    ResourceNotFoundError,
    ResourceNotFoundMalformedIdError,
)
from ajsonapi.events import DeleteEvent, UpdateEvent
from ajsonapi.exceptions import ErrorsException, NonexistentRelationshipIdError
from ajsonapi.id_value import IdValue
from ajsonapi.logging import log
from ajsonapi.responses import document_response, no_data_response
from ajsonapi.uri.collection import (
    Collection,
    VerifyRelationshipIdsEnvironment,
    parse_request_document_data,
)
from ajsonapi.uri.resource import Resource


class Object(Resource):
    """Class object represents a resource object."""

    def __init__(self, collection, id_):
        self.collection = collection
        self.id = id_  # pylint: disable=invalid-name
        self.table = collection.table
        self.pool = self.table.pool

    async def parse_document(self, request):
        """Parses the request document for a valid resource object.

        Args:
            request (Request): the request to be parsed.

        Returns:
            A ResourceObjectDocument.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource object.
        """

        data = await parse_request_document_data(request)
        document = ResourceObjectDocument(data)
        document.parse(self.collection, self.id.name)  # Needs an id_name
        return document

    async def get(self, query):
        """Produces the response for a GET /{collection}/{id} request."""

        links = {'self': self.path()}
        return document_response(await self.to_document(query=query,
                                                        links=links))

    async def delete(self):
        """Removes the object from the database and produces the response for
        a DELETE /{collection}/{id} request.
        """

        stmt = f'DELETE FROM {self.table.name} WHERE id = $1'
        try:
            async with self.table.pool.acquire() as connection:
                log.debug(f'{stmt}: {self.id.number}')
                result = await connection.execute(stmt, self.id.number)
        except ForeignKeyViolationError:
            raise ErrorsException(
                [DeleteObjectRemotelyRelatedError(self.path())])
        if result == 'DELETE 0':
            raise ErrorsException([ResourceNotFoundError(self.path())])
        event = DeleteEvent({
            'type': self.collection.name,
            'id': str(self.id.number),
            'links': {
                'self': f'/{self.collection.name}/{self.id.number}'
            },
        })
        await event.broadcast()
        return no_data_response()

    async def patch(self, document, query):
        """Patches the object in the database and produces the response for a
        PATCH /{collection}/{id} request.
        """

        if document.attributes or document.relationships:
            env = PatchObjectEnvironment(self, document.attributes,
                                         document.relationships, query)
            await env.run()
        links = {'self': self.path()}
        document = await self.to_document(query=query, links=links)
        event = UpdateEvent(document['data'])
        await event.broadcast()
        return document_response(document)

    async def patch_task(self, env):
        """Update the database collection table based on the PATCH
        /{collection}/{id} request.
        """
        local_relationships = {
            self.table.lfkey_by_relationship_name[name]: value
            for name, value in env.relationships.items()
            if name in self.table.lfkey_by_relationship_name
        }
        lfkeys = local_relationships.keys()
        lfkey_values = [id_.number for id_ in local_relationships.values()]
        column_names = chain(env.attributes.keys(), lfkeys)
        column_values = ', '.join(
            f'{col} = ${index+2}' for index, col in enumerate(column_names))
        stmt = (f'UPDATE {self.table.name} '
                f'SET {column_values} '
                f'WHERE id = $1')
        try:
            async with env.lock:
                async with env.connection.transaction():
                    log.debug(f'{stmt}: {env.attributes.values()}, '
                              f'{self.id.number}')
                    result = await env.connection.execute(
                        stmt, env.obj.id.number, *env.attributes.values(),
                        *lfkey_values)
            if result == 'UPDATE 0':
                env.cancel()
                raise ErrorsException([ResourceNotFoundError(self.path())])
        except ForeignKeyViolationError:
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def to_document(self, query=None, links=None):
        """Creates the document for this resource object.

        Args:
            links (dict): value for the links field in the response.
        """

        data, included = await self.to_data_included(query)
        document = {'data': data}
        if included is not None:
            document['included'] = included
        if links:
            document['links'] = links
        return document

    async def to_data_included(self, query):
        """Creates the response document's data and included member values."""

        env = GetObjectEnvironment(self, query)
        await env.run()
        return env.data, env.included

    async def get_object_data(self, env):
        """Creates the response data member for attributes and local
        relationships.

        Specifically, after this function call completes, the environment's
        data member contains something like the following object:

        {
            'type': 'centers',
            'id': '1',
            'attributes': {
                'attr_int': 1,
                'attr_str': 'one',
            },
            'relationships': {
                'one_one_local': {
                    'data': {
                        'type': 'one_one_locals',
                        'id': UUID_11,
                    },
                    'links': {
                        'self': '/centers/1/relationships/one_one_local',
                        'related': '/centers/1/one_one_local',
                    },
                },
                'many_one': {
                    'data': {
                        'type': 'many_ones',
                        'id': UUID_31,
                    },
                    'links': {
                        'self': '/centers/1/relationships/many_ones',
                        'related': '/centers/1/many_ones',
                    },
                },
            },
            'links': {
                'self': '/centers/1',
            },
        }
        """
        fields = env.query['fields']
        if self.collection in fields:
            columns = chain(['id'], (str(col)
                                     for col in self.table.columns
                                     if col in fields[self.collection]))
        else:
            columns = (str(col) for col in self.table.columns)
        # We could remove 'id' because we know it already.
        stmt = (f"SELECT {', '.join(columns)} "
                f"FROM {self.table.name} WHERE id = $1")
        async with env.lock:
            log.debug(f'{stmt}: {self.id.number}')
            record = await env.connection.fetchrow(stmt, self.id.number)
            if not record:
                env.cancel()
                raise ErrorsException([ResourceNotFoundError(self.path())])
        env.data = self.record_to_data(record)
        env.event.set()

    def record_to_data(self, record):
        """Converts a asyncpg record into a JSON API data member."""
        attributes = {}
        relationships = {}
        for col, value in record.items():
            if col == 'id':
                continue
            if col in self.table.attribute_names:
                attributes[col] = value
            else:  # col is a lfkey
                rel = self.table.relationship_by_lfkey[col]
                relationships[rel.name] = self.to_relationships(rel, value)
        data = {
            'type': self.collection.name,
            'id': self.id.name,
        }
        if attributes:
            data['attributes'] = attributes
        if relationships:
            data['relationships'] = relationships
        return data

    def path(self):
        """Creates the path to the collection."""

        return f'/{self.collection.name}/{self.id.name}'


class GetObjectEnvironment:
    """Class GetObjectEnvironment is the execution environment for a GET
    /{collection}/{id} request.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, obj, query):
        self.obj = obj
        self.query = query
        self.tasks = []
        self.connection = None
        self.lock = Lock()
        self.data_environment = None
        self.data = None
        self.included_environments = []
        self.included = None

    def create_data_environment(self):
        """Creates the environment for collecting the data member of the GET
        /{collection}/{id} response.
        """
        self.data_environment = GetObjectDataEnvironment(self)

    def create_included_environments(self, ids, includes):
        """Creates the environment for collecting the included member of the
        GET /{collection}/{id} response.
        """
        if includes is None:
            return
        for relationship, remote_includes in includes.items():
            remote_collection = relationship.rtable.collection
            remote_ids = relationship.object_remote_ids(ids)
            self.included_environments.append(
                GetObjectIncludedEnvironment(self.obj.id.number,
                                             remote_collection, remote_ids,
                                             self))
            if remote_includes:
                self.create_included_environments(remote_ids, remote_includes)

    def create_tasks(self):
        """Creates all the tasks to generate the GET /{collection}/{id}
        response.
        """

        self.create_data_environment()
        self.create_included_environments(None, self.query['includes'])
        self.tasks = [
            create_task(env.run()) for env in chain([self.data_environment],
                                                    self.included_environments)
        ]

    async def run(self):
        """Executes the GET /{collection}/{id} tasks."""

        async with self.obj.table.pool.acquire() as self.connection:
            async with self.connection.transaction():
                self.create_tasks()
                await gather(*self.tasks)
        self.data = self.data_environment.data
        if self.query['includes']:
            self.included = list(
                chain.from_iterable(env.included_by_id.values()
                                    for env in self.included_environments))


class GetObjectDataEnvironment:
    """Class GetCollectionDataEnvironment is the execution environment for the
    data part of the GET /{collection}/{id} response.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, env):
        self.obj = env.obj
        self.query = env.query
        self.tasks = []
        self.gather_future = None
        self.connection = env.connection
        self.lock = env.lock
        self.data = None
        self.event = Event()

    def create_tasks(self):
        """Creates all the tasks for querying the database to generate the
        data part of a GET /{collection}/{id} response.
        """

        self.tasks = [create_task(self.obj.get_object_data(self))]
        query_fields = self.query['fields']
        if self.obj.collection in query_fields:
            self.tasks.extend([
                create_task(rel.get_object_data(self))
                for rel in query_fields[self.obj.collection]
                if rel in self.obj.table.remote_relationships
            ])
        else:
            self.tasks.extend([
                create_task(rel.get_object_data(self))
                for rel in self.obj.table.remote_relationships
            ])

    async def run(self):
        """Executes the GET /{collection}/{id} tasks."""

        self.create_tasks()
        self.gather_future = gather(*self.tasks)
        await self.gather_future

    def cancel(self):
        """Cancels the environment."""

        self.gather_future.cancel()


class PatchObjectEnvironment:
    """Class PatchObjectEnvironment is the execution environment for a PATCH
    /{collection}/{id} request.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, obj, attributes, relationships, query):
        self.obj = obj
        self.attributes = attributes
        self.relationships = relationships
        self.query = query
        self.connection = None
        self.lock = Lock()
        self.tasks = []
        self.gather_future = None

    def create_tasks(self):
        """Creates all the tasks to execute the PATCH /{collection}/{id}
        request.
        """
        self.tasks = [create_task(self.obj.patch_task(self))]
        self.tasks.extend([
            create_task(rel.patch_object_task(self))
            for rel in self.obj.table.remote_relationships
            if rel.name in self.relationships
        ])

    async def run(self):
        """Executes the PATCH /{collection}/{id} tasks."""

        async with self.obj.table.pool.acquire() as self.connection:
            async with self.connection.transaction():
                try:
                    async with self.connection.transaction():
                        self.create_tasks()
                        self.gather_future = gather(*self.tasks)
                        await self.gather_future
                except NonexistentRelationshipIdError:
                    env = VerifyRelationshipIdsEnvironment(
                        self.obj.table, self.relationships, self.connection)
                    await env.run()

    def cancel(self):
        """Cancels the environment."""

        self.gather_future.cancel()


def parse(request):
    """Gets an object associated with the request.

    Args:
        request: Incoming Http(s) request.

    Exceptions:
        ErrorsException: Exception containing a 'resource not found' error.
    """

    collection_name = request.match_info['collection']
    id_name = request.match_info['id']

    try:
        collection = Collection.by_name[collection_name]
    except KeyError:
        raise ErrorsException([ResourceNotFoundError(f'/{collection_name}')])

    try:
        id_ = IdValue(id_name)
    except ValueError:
        raise ErrorsException(
            [ResourceNotFoundMalformedIdError(f'/{collection_name}/{id_name}')])

    return Object(collection, id_)
